#include "global.h"
#define epsilon 1e-12

#ifdef CG 
void preconditioner(double* A_old,int* CN_old, int* RS_old,int numberOfNodes,double* C)
{   
  C[0]=1/A_old[0];
  for(int i=1;i<numberOfNodes-1;i++)
  {
	for(int k=RS_old[i];k<RS_old[i+1];k++)
	  if(CN_old[k]<=i)
		C[i]=1/A_old[k];
	  else break;
  }
}

void PreCG(double *A_mat, int* CN_mat, int *RS_mat, int N,double *b,double *x,double *C)
{    
  double *r  = new double[N];
  double *h  = new double[N];
  double *d  = new double[N];
  if(!(d && h && r))
  {
	cout<<"Error: Memory allocation failed"<<endl;
	exit(1);
  }
  double delta_0=0;

  for(int i=0;i<N;i++)
	x[i]=C[i]*b[i];

  for(int i=0;i<N;i++)          /*for the ith row, this loop computes*/
  {                      /*multiplication from the start of the row to the diagonal element*/
	r[i]=0;
	for(int j=RS_mat[i];j<RS_mat[i+1];j++)
	  r[i]+=A_mat[j]*x[CN_mat[j]];
	r[i]-=b[i];
	h[i]=C[i]*r[i];    
	d[i]=-h[i];        
	delta_0+=r[i]*h[i];  
  } 
  if(delta_0<=epsilon) 
  { 	
	delete [] r;
	delete [] d;
	delete [] h;
	return;
  }
  double tow,beta,delta_1,tran_d_h=0;
  int  count=0;
  while(true)
  {
	tran_d_h=0;
	for(int i=0;i<N;i++) 
	{                  
	  h[i]=0;
	  for(int j=RS_mat[i];j<RS_mat[i+1];j++)
		h[i]+=A_mat[j]*d[CN_mat[j]];
	  tran_d_h+=d[i]*h[i];
	}
	tow = delta_0/tran_d_h;
	delta_1=0;
	for(int i=0;i<N;i++)
	{
	  x[i]+=tow*d[i];
	  r[i]+=tow*h[i];
	  h[i]=C[i]*r[i];
	  delta_1+=r[i]*h[i];
	}

	if(delta_1<=epsilon)
	{
	  cout<<"The solution has converged in "<<count<<" iterations"<<endl;
	  //	mat_vect(N,A_mat,CN_mat,RS_mat,b,x);
	  delete [] r;
	  delete [] d;
	  delete [] h;
	  return;
	}
	beta=delta_1/delta_0;
	delta_0=delta_1;
	for(int i=0;i<N;i++)
	{
	  d[i]*=beta;
	  d[i]-=h[i];
	}
	count++;
  }
}      

void GraphOfNetwork::pcg(double *A_old, double* b, int* CN_old, int *RS_old,
	int numberOfNodes,double* xOutPut)
{
  clock_t time2=clock();
  static double timeInMatSolve=0;
  double *C= new double[numberOfNodes-1];
  if(!C)
  {
	cout<<"Error:Memory allocation failed for C, Exiting abnormally!"<<endl;
	exit(1);
  }
  preconditioner(A_old,CN_old,RS_old,numberOfNodes,C);
  PreCG(A_old,CN_old,RS_old,numberOfNodes-1,b,xOutPut,C);
  delete [] C;
  clock_t time3=clock();
  timeInMatSolve+=double(time3-time2)/CLOCKS_PER_SEC;
  cout << "Time taken in Matrix Solving (CG) till this iteration is " << timeInMatSolve << endl;
}
#endif
