#include "global.h"
#ifdef MCF
#define TREE '^'
void GraphOfNetwork::buildFlowGraph(char *fgfile)
{
   int sourceNode=0,sinkNode=0;
   char file[80],tempStr[80];
   strcpy(file,fgfile);
   edge e;
   FILE *from,*fgnew;
   int flow=0,capacity=0,cost=0,i=0,j=-1,absFlow=0;
   int *excess=NULL;
   double diodeI=0;
   double offCurrentLimit=0.1*(1.0/scalecapacity);
   double TotalCost=0.0;  
   bool failed;
  // int tempCap,tempFlow;
   from=fopen(file,"r+");
   strcat(file,".fg");
   fgnew=fopen(file,"w+");
   if(from == NULL)
   {
        cout<<" Flow Circuit File Not Found " <<endl;
        exit(0);
   }
   e=edgePtrList[0];
   while (fgets(tempStr, 80, from) != NULL)
   {
       switch(tempStr[0])
       {
       case 'p':
         sscanf(tempStr,"p min %d %d",&_Nodes,&_Edges);
         fgNodePtrList = (fgnode*) calloc(_Nodes,sizeof(fgnode));
         fgEdgePtrList = (fgedge*) calloc(_Edges,sizeof(fgedge));
	 excess = (int*) calloc(_Nodes,sizeof(excess));
	 break;
       case 'n':
         sscanf(tempStr,"n %d %d",&sourceNode,&flow);
         sourceNode=nodeTable[sourceNode];
	 --sourceNode;
         excess[sourceNode]+=flow; 	 
         absFlow+=abs(flow); 
	 j++;
         break;
       case 'a':
         sscanf(tempStr,"a %d %d %*d %d %d",&sourceNode,&sinkNode,&capacity,&cost);
         sourceNode=nodeTable[sourceNode];
	 sinkNode=nodeTable[sinkNode];
	 --sourceNode; --sinkNode;
         if(fgNodePtrList[sourceNode] == NULL) fgNodePtrList[sourceNode] = new FGNode(sourceNode);
         if(fgNodePtrList[sinkNode] == NULL) fgNodePtrList[sinkNode] = new FGNode(sinkNode);
         if(capacity>absFlow)
	    capacity=absFlow;
	 diodeI=edgePtrList[j]->EdgeValue;
	 failed=false;
         if(edgePtrList[j+1]->devCharNotSatisfied ||edgePtrList[j+3]->devCharNotSatisfied)  //Erroreneous 
	    failed=true;
       else if(diodeI >= edgePtrList[j+5]->EdgeValue){                                    //Saturated
	    if(capacity>1000){
	    	excess[sourceNode]-=capacity;
	    	excess[sinkNode]+=capacity;
            	TotalCost+=(double)cost*(double)capacity;
	    	capacity=0;
	    }	
	 }
         else if(diodeI < offCurrentLimit){
	    diodeI=0;
	    if(capacity>1000)
	    	capacity=1000;
	 }
/*	 else{
	    diodeI*=scalecapacity;
	    if(diodeI<1000){
	       tempCap=(int)diodeI+1000;
	       if(tempCap<capacity)
	       capacity=tempCap;
	    }
	    else{
	       tempCap=(int)diodeI+100;
               if(tempCap<capacity)
               	  capacity=tempCap;
	       tempFlow=(int)diodeI-100;
               excess[sourceNode]-=tempFlow;
	       excess[sinkNode]+=tempFlow;
               TotalCost+=(double)cost*(double)tempFlow;
            }  
	 }*/	   	
         if(capacity==0)
		--_Edges;
	 else
         	addFgEdge(FlowEdge,fgNodePtrList[sourceNode],fgNodePtrList[sinkNode],cost,diodeI,capacity,i,failed);
	 j+=6;
         e=edgePtrList[j];
         break;
       case 'c' :
         break;
       default : cout <<"File doesn't looks to be in DIMACS format " << endl;exit(0);
       }
  }
  fclose(from);
  free(edgePtrList);
  bool* nodeCovered = (bool*) calloc(_Nodes,sizeof(bool));
  queue <fgnode> queueOfNodes;
  fgnode predecessor,n;
  fgedge e1;
  _Nodes=1;  
  forall_fgedges(e1)
	e1->edgeCovered=false;
  forall_fgnodes(n){
     if(nodeCovered[n->nodeNumber])
	continue;	
     queueOfNodes.push(n);
     nodeCovered[n->nodeNumber]=true;	
     while(!queueOfNodes.empty())
     {
         predecessor=queueOfNodes.front();
         queueOfNodes.pop();
         forall_adj_fgedges(e1,predecessor)
         {
            n=other_fgnode(e1,predecessor);
            if(!nodeCovered[n->nodeNumber])
            {
               nodeCovered[n->nodeNumber]=true;
               queueOfNodes.push(n);
               e1->edgeCovered=true;	
               ++_Nodes;
            }
         }
     }
  }
  cout<<"Number of Nodes: "<<_Nodes<<endl;
  if(fgnew == NULL){
	cout<<" Error in creating Flow Graph file" <<endl;
	exit(0);
  } 
  fprintf(fgnew,"c Reduced MinCostFlow problem\n");
  fprintf(fgnew,"p min %d %d\n",_Nodes,_Edges);
  for(int i=0;i<_Nodes;++i)
     if(excess[i])
	fprintf(fgnew,"n %d %d\n",reverseNodeTable[i+1],excess[i]);			
  for(int i=0;i<_Edges;++i){
	e1=fgEdgePtrList[i];
       	fprintf(fgnew,"a %d %d 0 %d %d\n",reverseNodeTable[e1->Source->nodeNumber+1],reverseNodeTable[e1->Sink->nodeNumber+1],e1->Capacity,(int)e1->Voltage);
  }		  
  fclose(fgnew);
  printf("Total Cost: %lf\n",TotalCost);
  FILE *fpCost = fopen(".mcfCostdcAnaIITB","w");
  fprintf(fpCost,"%lf\n",TotalCost);
  fclose(fpCost);
}

void GraphOfNetwork::addFgEdge(char Type,fgnode Source,fgnode Sink,double depValue,double Value,int cap,int& i,char status)
{
  fgEdgePtrList[i] = new FGEdge(Value,Type,Source,Sink);
  if(Type==VoltageSource)
        fgEdgePtrList[i]->EdgeValue=depValue;
  else
        fgEdgePtrList[i]->Voltage=depValue;
  fgEdgePtrList[i]->Capacity=cap;
  fgEdgePtrList[i]->edgeCovered=status;
  Source->addEdgeList(fgEdgePtrList[i]);
  Sink->addEdgeList(fgEdgePtrList[i]);
  ++i;
}

#endif

