//Pre-Congugate Gradient Methods
// M. Sudarshan Reddy
// M. Tech(Control& Computing)
// Electrical Engg
// IIT Bombay, Mumbai-400076
// Guide:Prof. H. Narayanan


#include"global.h"
#ifdef CG  
#define ACCEPTABLE_ERROR_NORM 1e-07
#define MAX_ITERATIONS 10000

void MatrixVectorMul(double *A,int *CN,int *RS,double *b,int N,double *c)
{
  int k,l;
  for(int i=0;i<N;i++)
  {	
	c[i]=0;
	for(l=RS[i];l<RS[i+1];l++)
	{
	  k=CN[l];
	  c[i]+=A[l]*b[k];
	  if(i!=k)
		c[k]+=A[l]*b[i];
	}	
  }
}
double Dotproduct(double *x,double *y,int n)
{
  double d=0;
  for(int i=0;i<n;i++)
	d+=x[i]*y[i];
  return(d);
}
void CGwithdiagonalprec(double *A,int *CN,int *RS,double *b,double *result,int N)  
{     
  printf("DPCG called to solve Ax=b\n");
  clock_t start1, stop1;
  int i,iterationCount=0;
  double *residue,*searchDirec,*h,olddelta,newdelta,t,beta,*c;
  double *prec;
  h=(double*)malloc(sizeof(double)*(N));
  c=(double *)malloc(N*sizeof(double));
  residue=(double *)malloc(N*sizeof(double));
  searchDirec=(double *)malloc(N*sizeof(double));
  prec=(double *)malloc(N*sizeof(double));
  if(!c||!h||!prec||!searchDirec||!residue)
  {
	printf("Memory allocation failed in DPCG Sorry!\nAbnormal termintion\n");
	exit(1);
  }
  //***************Diagonal Preconditioner Extraction*************
/*
for(i=0;i<RS[N];i++) 
 {
 if(A[i] < 1e-6) 
 cout << A[i] << endl; 
}
*/
  for(i=0;i<N;i++)
  {	
	prec[i]=(double)1/A[RS[i+1]-1];
	if(A[RS[i+1]-1]<1e-12)
	{
          cout<<" RS " << RS[i+1]-1<<endl; 
          cout<< i <<" "<<N<< endl ; 
          cout<< A[-1] << endl; 
	  cout<<"There is a zero "<< A[RS[i+1]-1]<<" diagonal entry I can't help you..........\n";
	  exit(1);
	}
  }
  //********************Initialization*******************	
  for(i=0;i<N;i++)
	result[i]=prec[i]*b[i];
  MatrixVectorMul(A,CN,RS,result,N,c);
  for(i=0;i<N;i++)
  {
	residue[i]=c[i]-b[i];
	h[i]=residue[i]*prec[i];
	searchDirec[i]=-h[i];
  }
  olddelta=Dotproduct(residue,h,N);
  start1=clock(); 
  //********************** The Main Algorithm************	
  while(fabs(sqrt(olddelta))>ACCEPTABLE_ERROR_NORM)
  {	
	iterationCount++;
	if(iterationCount==MAX_ITERATIONS)
	{
	  cout<<"Sorry I am not able to converge to the solution\n";
	  cout<<"Iterations exceeded the Maximum limit" <<MAX_ITERATIONS<<endl;
	  exit(1);
	}
	MatrixVectorMul(A,CN,RS,searchDirec,N,h);
	for(i=0,t=0;i<N;i++)
	  t+=searchDirec[i]*h[i];
	t=(double)olddelta/t;
	for(i=0;i<N;i++)
	{
	  result[i]+=t*searchDirec[i];
	  residue[i]+=t*h[i];
	}
	for(i=0;i<N;i++)
	  h[i]=prec[i]*residue[i];
	newdelta=Dotproduct(residue,h,N);		
	beta=(double)newdelta/olddelta;
	for(i=0;i<N;i++)
	  searchDirec[i]=-h[i]+beta*searchDirec[i];
	olddelta=newdelta;
  }
  //******************CG completed the solution***************
  printf("DPCG completed the solution process\n");
  cout<<"Second norm of residue:"<<sqrt(olddelta)<<endl;
  stop1=clock();
  printf("total time taken by DPCG(inner clock):%lf\n",(double)(stop1-start1)/CLOCKS_PER_SEC);
  MatrixVectorMul(A,CN,RS,result,N,c);
  for(i=0;i<N;i++)
	residue[i]=c[i]-b[i];
  t=(double)Dotproduct(residue,residue,N)/Dotproduct(result,result,N);
  fprintf(stdout,"Matrix size:%d\n",N);
  fprintf(stdout,"No of iterations:%d\n",iterationCount);
  fprintf(stdout,"Ratio of error norm to the solution norm:%e\n",sqrt(t));
	free(c);
	free(h);
	free(prec);
	free(searchDirec);
	free(residue);
}	

void symmetricmatrixconverter(double *A_old,int *CN_old,int *RS_old,int numberOfNodes,int totalSparseElement)
{     
	int P,k,m;   
        double *A;
	int *CN,*RS;
        A  = new double[((numberOfNodes-1)+totalSparseElement)/2];
	CN = new int[((numberOfNodes-1)+totalSparseElement)/2];
	RS = new int[numberOfNodes];
	for(int i=0;i<numberOfNodes;i++)
	   RS_old[i]=RS_old[i]+1;
	A[0] = A_old[0];
	CN[0] = CN_old[0];
        RS[0] = 0;
	RS[1] = 1;
	int j=1;
	for(int row=1;row<numberOfNodes-1;row++)
        {
                P=RS_old[row+1]-RS_old[row];
		int numOfelmrow =0;
		for(int l=0;l<P;l++)
	        {
                    k=RS_old[row]+l;
		    m=CN_old[k-1];
		    if(m<=row)
		    {
		        A[j]=A_old[k-1];
		        CN[j]=CN_old[k-1]; 
			j++;
                        numOfelmrow++;
		    }
			 else break;
	       }	   
		 RS[row+1]=RS[row]+numOfelmrow;
	 }
	A_old=(double *)realloc(A_old,sizeof(double)*((numberOfNodes-1)+totalSparseElement)/2);
	CN_old=(int *)realloc(CN_old,sizeof(int)*((numberOfNodes-1)+totalSparseElement)/2);
	if(!A_old||!CN_old)
	{
		cout<<"Sorry Reallocation failed in Converter\n";
		exit(1);
	}  
	for(int i=0;i<((numberOfNodes-1)+totalSparseElement)/2;i++)
		 A_old[i]=A[i];
        for(int i=0;i<((numberOfNodes-1)+totalSparseElement)/2;i++)
	  	CN_old[i]=CN[i];
	for(int i=0;i<numberOfNodes;i++)
	  	 RS_old[i]=RS[i];
         delete []A;	 
         delete []CN;
	 delete []RS;
     }

struct listofcolindices
{
	int index;
	float value;
	struct  listofcolindices *next;
};
typedef struct listofcolindices linked_list;
void incompleteCholesky(double *A,int *CN,int *RS,double *L,double *U,int *CN1,int *RS1,int N)
{
	int i,p,j,l,k,n,m1,m2,test1,test2,temp,m,q;
      	double l1;
	clock_t start1,stop1;
	linked_list *list[N],*head[N];
	int *temparray=new int[N];
	for(i=0;i<N;i++)
	{
		list[i]=new linked_list[1];
		head[i]=list[i];
		temparray[i]=0;
	}	
	start1=clock();
      	for(i=0;i<N;i++)
      	{
        	p=RS[i+1]-RS[i];
              	for(l=0;l<p;l++)
	   	{		
                	j=RS[i]+l;
                     	k=CN[j];
	      	     	q=RS[k+1]-RS[k];	
		     	temp=(p>q)?p:q;temp--;	
		     	temp=(k>temp)?temp:k;
                     	l1=0;test1=0;test2=0;
		     	if(i!=k)
              		for(n=0;n<temp;n++)
			{	
				m1=CN[RS[i]+test1];
				m2=CN[RS[k]+test2];
				if(m2<m1)
				{
					test2++;continue;
				}
                		if(m2==m1)
				{       
					l1+=L[RS[i]+test1]*L[RS[m1+1]-1]*L[RS[k]+test2];
		                        test1++;test2++;continue;
				}
				if(m2>m1) test1++;
				
             		 }
             		else
             		for(n=0;n<p-1;n++)
			{	m=RS[i]+n;
				l1+=L[m]*L[m]*L[RS[CN[m]+1]-1];
			}
              		if(i==k)
               		L[j]=A[j]-l1;
              		else
              		L[j]=(double)(A[j]-l1)/L[RS[k+1]-1];
			list[k]->index=i;
			list[k]->value=L[j];
			list[k]->next=new linked_list[1];
			list[k]=list[k]->next;
			temparray[k]++;
                }
        }
	for(i=0;i<N;i++)
	{
	  	list[i]->next=NULL;
		list[i]=head[i];
	}
	RS1[0]=0;
	for(i=1;i<N+1;i++)
	RS1[i]=RS1[i-1]+temparray[i-1];
	for(i=0;i<N;i++)
	{
		p=RS1[i+1]-RS1[i];
		for(l=0;l<p;l++)
		{	j=RS1[i]+l;
			CN1[j]=list[i]->index;
			U[j]=list[i]->value;
			list[i]=list[i]->next;
		}
	}
	stop1=clock();
  	cout<<"Time taken by the incomplete LDL' factorization:"<<(double)(stop1-start1)/CLOCKS_PER_SEC<<endl;
	for(i=0;i<N;i++)
  	while(head[i]->next)
     	{	list[i]=head[i]->next;
	 	free(head[i]);
		head[i]=list[i];
	}
  	for(i=0;i<N;i++)
  	free(head[i]);
	delete []temparray;
}

void Sorttheadjacencyset(int *adjacencyset, int count,int *RS)
{	
	int i,j,q,temp;
	int p;	
	for(i=0;i<count-1;i++)
	{
                p=RS[adjacencyset[i]+1]-RS[adjacencyset[i]];
		for(j=i+1;j<count;j++)
		{	
			q=RS[adjacencyset[j]+1]-RS[adjacencyset[j]];
                	if(p>q)
                	{
                        	temp=adjacencyset[j];
				adjacencyset[j]=adjacencyset[i];
				adjacencyset[i]=temp;
                	}
		}
	}
}
 	
void cuthillMckee(double *A,int *CN,int *RS,int N,double *b,int *dperm)
{
	cout<<"CM routine is called\n";
	int i,p=0,m=0,mindegree=0,l,j,k;
	int *adjacencyset,count=0,nodeordercount=0;
	char color[N];
	double temp1;
	clock_t stop,start;
	int *perm=new int[N];
	start=clock();
	for(i=0;i<N;i++)
		color[i]='B';//Initially all nodes will be black
// **********************Hunt for unity degree nodes***************	
	for(i=0;i<N;i++)
	{	
		if((RS[i+1]-RS[i])==1)
		{
			perm[nodeordercount++]=i;
			color[i]='W';
		}
	}
	mindegree=N;m=N;
// *********************Min degree node among the remaining nodes ******
	for(i=0;i<N;i++)
	{
		p=RS[i+1]-RS[i];
		if(mindegree>p&&color[i]=='B')	
		{
			m=i;
			mindegree=p;
		}
	}
	perm[nodeordercount++]=m;
	color[m]='W';//Make the starting node white
	adjacencyset=new int[mindegree];count=0;
	for(l=0;l<mindegree;l++)
	{
		j=RS[m]+l;
		k=CN[j];
		if(color[k]=='B')
		{
			color[k]='R';
			adjacencyset[count++]=k;
		}
	}
//**********************Actual algo**************************************
	while(count)
	{	
//		cout<<"adjacency list"<<endl;
//		for(i=0;i<count;i++)
//			cout<<adjacencyset[i]<<" ";cout<<endl;
		Sorttheadjacencyset(adjacencyset,count,RS);
//		for(i=0;i<count;i++)
//			cout<<adjacencyset[i]<<" ";cout<<endl;
		for(i=0;i<count;i++)
		{
			perm[nodeordercount++]=adjacencyset[i];
			color[adjacencyset[i]]='W';
		}i=nodeordercount-count;
		for(count=0;i<nodeordercount;i++)
		{	
			p=RS[perm[i]+1]-RS[perm[i]];
			adjacencyset=(int *)realloc(adjacencyset,sizeof(int)*(count+p));
			if(adjacencyset==NULL)
			{
				cout<<"Sorry!!Reallocation of memorey failed in the CM routine\n";
				exit(1);
			}
			for(l=0;l<p;l++)
			{	j=RS[perm[i]]+l;
				k=CN[j];
				if(color[k]=='B')
				{
					adjacencyset[count++]=k;
					color[k]='R';
				}
			}
		}
		if(count==0&&nodeordercount<N)
		for(i=0;i<N;i++)
		{	
			if(color[i]=='B')
			{
				adjacencyset=(int *)realloc(adjacencyset,sizeof(int)*(1));
				adjacencyset[count++]=i;
				i=N;
			}
		}
	}		
	double *B=new double[RS[N]];
	double *c=new double[N];
	int *RS1=new int[N+1];
	int *CN1=new int[RS[N]];RS1[0]=0;
//	cout<<"nodeordercount: "<<nodeordercount<<" "<<N<<endl;
//	for(i=0;i<N;i++)
//		cout<<perm[i]<<" ";cout<<endl;
	//cout<<"RCM ordering\n";
	/*for(i=0;i<N/2;i++)
	{
		temp=perm[i];
		perm[i]=perm[N-i-1];
		perm[N-i-1]=temp;
	}*/
	for(i=0;i<N;i++)
		dperm[perm[i]]=i;
	for(i=0;i<N;i++)
	{
		p=RS[perm[i]+1]-RS[perm[i]];
		RS1[i+1]=RS1[i]+p;
		for(l=0;l<p;l++)
		{
			j=RS[perm[i]]+l;
			CN1[RS1[i]+l]=dperm[CN[j]];
			B[RS1[i]+l]=A[j];
		}
	}
	for(i=0;i<N;i++)
	{
		p=RS1[i+1]-RS1[i];
		for(l=0;l<p-1;l++)
		for(m=l+1,k=CN1[RS1[i]+l];m<p;m++)
		if(k>CN1[RS1[i]+m])
		{
			CN1[RS1[i]+l]=CN1[RS1[i]+m];
			CN1[RS1[i]+m]=k;
			k=CN1[RS1[i]+l];
			temp1= B[RS1[i]+l];
			B[RS1[i]+l]=B[RS1[i]+m];
			B[RS1[i]+m]=temp1;
		}
	}
	for(i=0;i<N;i++)
	{
		p=RS1[i+1]-RS1[i];
		for(l=0;l<p;l++)
		{
			j=RS1[i]+l;
			k=CN1[j];
			A[j]=B[j];
			CN[j]=CN1[j];
		}
		RS[i+1]=RS1[i+1];
	}
	/*
	FILE *fp=fopen("check1","w");
	for(i=0;i<N;i++)
	{
		p=RS[i+1]-RS[i];
		for(l=0;l<p;l++)
		{
			j=RS[i]+l;
			k=CN[j];
			fprintf(fp,"%d %d %lf\n",i,k,A[j]);
		}
	}
	*/
	for(i=0;i<N;i++)
		c[i]=b[perm[i]];
	for(i=0;i<N;i++)
		b[i]=c[i];
	stop=clock();
	cout<<"Time taken="<<(double)(stop-start)/CLOCKS_PER_SEC;cout<<endl;
	delete []RS1;delete []B;delete []CN1;delete []c;delete[]perm;delete []adjacencyset;
}
 	
void Solvepreconditioner(double *L,int *CN,int *RS,double *U,int *CN1,int *RS1,double *h,int N)
{
        int i,j,p,k,l;
	for(i=0;i<N;i++)
        {		
                p=RS[i+1]-RS[i];
                for(l=0;l<p;l++)
                {
                        j=RS[i]+l;
                        k=CN[j];
                        if(i!=k)
                        h[i]-=h[k]*L[j];
                }
        }
		for(i=0;i<N;i++)
		  h[i]=(double)h[i]/L[RS[i+1]-1];
        for(i=N-1;i>=0;i--)
        {
                p=RS1[i+1]-RS1[i];
                for(l=p-1;l>0;l--)
                {
                        j=RS1[i]+l;
                        k=CN1[j];
                        if(i!=k)
                        h[i]-=h[k]*U[j];
                }
        }
        
     
}
void CGwithincompleteLDLTprec(double *A,int *CN,int *RS,double *b,double *result,int N)  
{
	cout<<"CGwith incomplete cholesky preconditioner called to solve Ax=b\n";
	clock_t start, stop;
	int nnz=RS[N];
	int i,iterationCount=0,*RS1,*CN1;
	double *residue,*searchDirec,*h,olddelta,newdelta,t,beta,*c,*L,*U;
	CN1=(int *)malloc(nnz*sizeof(int));
	c=(double *)malloc(N*sizeof(double));
	residue=(double *)malloc(N*sizeof(double));
    	searchDirec=(double *)malloc(N*sizeof(double));
    	h=(double *)malloc(N*sizeof(double));
    	L=(double *)malloc(nnz*sizeof(double));
    	U=(double *)malloc((nnz)*sizeof(double));
	RS1=(int *)malloc((N+1)*sizeof(int));
	if(!c||!h||!searchDirec||!residue||!CN1||!RS1||!L||!U)
        {
                printf("Memory allocation failed in IC-PCG Sorry!\nAbnormal termintion\n");
                exit(1);
        }
	start=clock();
	incompleteCholesky(A,CN,RS,L,U,CN1,RS1,N);
	stop=clock();
	fprintf(stdout,"Time taken to factorize:%lf\n",(double)(stop-start)/CLOCKS_PER_SEC);
	start=clock();
//********************Initialization*******************	
	for(i=0;i<N;i++)
		result[i]=b[i];
	Solvepreconditioner(L,CN,RS,U,CN1,RS1,result,N);
	MatrixVectorMul(A,CN,RS,result,N,c);
	for(i=0;i<N;i++)
		residue[i]=c[i]-b[i];
	for(i=0;i<N;i++)
	  h[i]=residue[i];
	Solvepreconditioner(L,CN,RS,U,CN1,RS1,h,N);
	for(i=0;i<N;i++)
		searchDirec[i]=-h[i];
	olddelta=Dotproduct(residue,h,N);
//********************** The Main Algorithm************	
	while(fabs(sqrt(olddelta))>ACCEPTABLE_ERROR_NORM)
	{	
		iterationCount++;
		if(iterationCount==MAX_ITERATIONS)
		{
			cout<<"Sorry I am not able to converge to the solution\n";
			cout<<"No Of iteration performed are:"<<MAX_ITERATIONS<<"\n";
			exit(1);
		}
		MatrixVectorMul(A,CN,RS,searchDirec,N,h);
		for(i=0,t=0;i<N;i++)
			t+=searchDirec[i]*h[i];
		t=(double)olddelta/t;
		for(i=0;i<N;i++)
		{
			result[i]+=t*searchDirec[i];
			residue[i]+=t*h[i];
		}
		for(i=0;i<N;i++)
		  h[i]=residue[i];
		Solvepreconditioner(L,CN,RS,U,CN1,RS1,h,N);
		newdelta=Dotproduct(residue,h,N);		
		beta=(double)newdelta/olddelta;
		for(i=0;i<N;i++)
			searchDirec[i]=-h[i]+beta*searchDirec[i];
		olddelta=newdelta;
	}
//******************CG completed the solution***************
 	printf("IC-PCG completed the solution process\n");
	cout<<"Second norm of residue:"<<sqrt(olddelta)<<endl;
	printf("Matrix size:%d\n",N); 
	stop=clock();
        printf("total time taken by IC-PCG(inner clock):%lf\n",(double)(stop-start)/CLOCKS_PER_SEC);
	MatrixVectorMul(A,CN,RS,result,N,c);
	for(i=0;i<N;i++)
		{	
	  		residue[i]=c[i]-b[i];
		}
	t=(double)Dotproduct(residue,residue,N)/Dotproduct(result,result,N);
	fprintf(stdout,"No of iterations:%d\n",iterationCount);
	fprintf(stdout,"Ratio of error norm to solution norm:%e\n",sqrt(t));
	free(c);
	free(h);
	free(searchDirec);
	free(residue);
}	

void GraphOfNetwork::snPCG(double *A,double *b,int *CN,int *RS,int N,double *result,bool choice)
{
	if(!choice)
		CGwithdiagonalprec(A,CN,RS,b,result,N-1);
		else 
		CGwithincompleteLDLTprec(A,CN,RS,b,result,N-1);
}

#endif 
