#include "global.h"
#define TREE '#'
#define OUTLOOP '@'
#define INLOOP '$'
#define BLANK '^'
void FlowGraph::findLoopOfUnsaturatedEdges()
{
  queue <node> queueOfNodes;
  stack<node> unsaturatedTree;
  node firstNode=nodePtrList[0];
  firstNode->potential = 0; 
  bool* nodeCovered = (bool*) calloc(_Nodes,sizeof(bool));
  node n1,n2; 
  edge e;
  int temp_cost;
  bool dir,unsatEdgeFound=false;
//This loop builds forest of unsaturated edges using bfs and find node potential by assigning 0 volt to datum node of each unsaturated tree.
  forall_nodes(firstNode)
  {
     if(nodeCovered[firstNode->nodeNumber]==true) continue; 
     if(firstNode->edgeList.empty()) continue;
     nodeCovered[firstNode->nodeNumber] =true;
     queueOfNodes.push(firstNode);
     while (!queueOfNodes.empty())
     {
           n1 = queueOfNodes.front();
     	   unsatEdgeFound=false;
           unsaturatedTree.push(n1);
	   forall_adj_edges(e,n1)
           {
     	   if(e->Source == e->Sink) continue;
             if(e->isEdgeClass(UNSAT))
             {
           	dir = (n1==e->Sink);
           	n2 = dir ? e->Source:e->Sink; 
           	if(!nodeCovered[n2->nodeNumber])
           	{
           	  unsatEdgeFound=true;
		  unsaturatedTree.push(n2);
           	  e->edgeCovered=TREE;
           	  if(dir)
           		n2->potential=n1->potential + e->EdgeCost;
           	  else
           		n2->potential=n1->potential - e->EdgeCost;
           	  nodeCovered[n2->nodeNumber]=true;
                  queueOfNodes.push(n2);
           	}
             }
           }
           queueOfNodes.pop();
     	   if(unsatEdgeFound!=true)
		unsaturatedTree.pop();
     }
  }
  bool unsatLoopFound;
  bool commonNodeFound=true;
  node nextNode=nodePtrList[0];                       //for only initialization
  if(!unsaturatedTree.empty())
 	nextNode=unsaturatedTree.top();
  node preNode,commonNode;

/*This loop finds f-circuits of Unsaturated edges which satify KVL. The cotree edge associated with the f-circuit is replaced by current source and tree edges associated with the f-circuit are replaced by voltage source */

   while(!unsaturatedTree.empty())
   {
     n1=unsaturatedTree.top();
     unsatLoopFound=false;
     forall_adj_edges(e,n1){
        if(e->Source==e->Sink) continue;
        if(e->EdgeClass==UNSAT && e->edgeCovered!=TREE && e->edgeCovered!=OUTLOOP && e->edgeCovered!=INLOOP){
           n2=other_node(e,n1);
 	   if(n1==e->Source)
                temp_cost=n1->potential-n2->potential;
           else
                temp_cost=n2->potential-n1->potential;
           if(e->EdgeCost==temp_cost){
		e->edgeCovered=INLOOP;
		e->EdgeType=CurrentSource;
		++_SimplifiedEdges;
		commonNodeFound=false;
                unsatLoopFound=true;
           }
        }
        if(e->edgeCovered==INLOOP && unsatLoopFound==false)
		unsatLoopFound=true;
     }      
     forall_adj_edges(e,n1)
        if(unsatLoopFound!=true){
       	   if(e->edgeCovered==TREE)
                e->edgeCovered=OUTLOOP;
        }
        else{
       	   if(e->edgeCovered==TREE && commonNodeFound==false){
		e->edgeCovered=INLOOP;
                e->EdgeType=VoltageSource;
		++_SimplifiedEdges;
		preNode=nextNode;
                nextNode=other_node(e,n1);
		if(preNode==nextNode){
			commonNodeFound=true;
			commonNode=preNode;
		}
           }
        } 
     unsaturatedTree.pop();
  }
/*  fstream out("circuit1",ios::out);
  out<<"c Modified MinCostFlow problem"<<endl;
  out<<"p min "<<_Nodes<<" "<<_Edges<<" "<<_SimplifiedEdges<<endl;
  out<<"n 1 "<<nodeCurrent<<endl;
  out<<"n "<<_Nodes<<" "<<-nodeCurrent<<endl;
  for(int i=0; i<_Edges;++i)
  {
    e=edgePtrList[i];
    if(e->isEdgeType(VoltageSource))
       out<<e->EdgeType<<" "<<e->Source->nodeNumber+1<<" "<<e->Sink->nodeNumber+1<<" "<<e->EdgeCost<<"  "<<endl;
    else if(e->isEdgeType(CurrentSource))
       out<<e->EdgeType<<" "<<e->Source->nodeNumber+1<<" "<<e->Sink->nodeNumber+1<<" "<<e->EdgeCurrent<<" "<<e->EdgeCost<<endl;
    else
       out<<e->EdgeType<<" "<<e->Source->nodeNumber+1<<" "<<e->Sink->nodeNumber+1<<" 0 "<<e->EdgeCapacity<<" "<<e->EdgeCost<<endl;
  } */ 
}

void FlowGraph::findCurrentSources()
{
  for(int i=0; i<_Edges;++i)
  {
    edge e=edgePtrList[i];
    if(e->EdgeClass==CurrentSource)
	e->EdgeType=CurrentSource;
  }
}	

void FlowGraph::replaceSaturatedEdges()
{
  edge e;
  node startNode=nodePtrList[0];
  node n1,n2;
  int* nodeBlockA = (int*) calloc(_Nodes,sizeof(int));
  bool* nodeCovered = (bool*) calloc(_Nodes,sizeof(bool));
  queue <node> queueOfNodes;
  int counter=1;
  bool treeEdgeFound,connectedBlockFound;
  list<int> tree_counter[_Nodes];
  int block1,block2;
  forall_nodes(startNode)
  {
     if(nodeCovered[startNode->nodeNumber]==true) continue; 
     if(startNode->edgeList.empty()) continue;
     queueOfNodes.push(startNode);
     treeEdgeFound=false;
     //This part builds tree of Unsaturated and Off edges(not forest)
     while(!queueOfNodes.empty())
     {
       n1=queueOfNodes.front();
       forall_adj_edges(e,n1)
       {
        n2= other_node(e,n1);
        if(n1==n2) continue;
        if(e->EdgeClass != SAT && e->EdgeType!=CurrentSource)
        {
          if(!nodeCovered[n2->nodeNumber])
          {
            e->edgeCovered = TREE;
            nodeCovered[n2->nodeNumber]=true;
            nodeBlockA[n2->nodeNumber] = counter;
            treeEdgeFound=true;
            queueOfNodes.push(n2);
          }
        }
       }
       queueOfNodes.pop();
     }
     if(treeEdgeFound)
     {
       nodeCovered[startNode->nodeNumber] =true;
       nodeBlockA[startNode->nodeNumber] = counter;
       ++counter;
     }
  }
  for(int i=0; i<_Edges;++i)
  {
    e=edgePtrList[i];
    if(e->EdgeClass!=SAT)
	continue;
    n1=e->Source;
    n2=e->Sink;
    
    if(!nodeCovered[n2->nodeNumber] &&!nodeCovered[n1->nodeNumber])
    {
        e->edgeCovered = TREE;
        nodeBlockA[n2->nodeNumber]=counter;
        nodeBlockA[n1->nodeNumber]=counter;
        nodeCovered[n1->nodeNumber] =true;
        nodeCovered[n2->nodeNumber] =true;
        ++counter;
        continue;
    }
    if(!nodeCovered[n2->nodeNumber]){
        e->edgeCovered = TREE;
        nodeBlockA[n2->nodeNumber]=nodeBlockA[n1->nodeNumber];
        nodeCovered[n2->nodeNumber] =true;
        continue;
    }
    if(!nodeCovered[n1->nodeNumber]){
        e->edgeCovered = TREE;
        nodeBlockA[n1->nodeNumber]=nodeBlockA[n2->nodeNumber];
        nodeCovered[n1->nodeNumber] =true;
        continue;
    }
    if(nodeBlockA[n2->nodeNumber]!=nodeBlockA[n1->nodeNumber])
    {
        connectedBlockFound=false;
        block1=nodeBlockA[n2->nodeNumber]<nodeBlockA[n1->nodeNumber]? nodeBlockA[n2->nodeNumber]:nodeBlockA[n1->nodeNumber];
        block2=nodeBlockA[n2->nodeNumber]<nodeBlockA[n1->nodeNumber]? nodeBlockA[n1->nodeNumber]:nodeBlockA[n2->nodeNumber];
	list<int> :: iterator p;
        while(!tree_counter[block2].empty()){
           p=tree_counter[block2].begin();
           if(*p==0)
                if(block1<*(++p))
                        block2=*p;
                else{
                        block2=block1;
                        block1=*p;
                }
           else
                break;
        }
        while(!tree_counter[block1].empty()){
           p=tree_counter[block1].begin();
           if(*p==0)
                block1=*(++p);
           else
                break;
        }
        for(p=tree_counter[block1].begin(); p!=tree_counter[block1].end(); ++p)
           if(*p==block2){
                connectedBlockFound =true;
                break;
           }
        if(block2==1 || block1==block2)
           continue;
        if(!connectedBlockFound){
           e->edgeCovered = TREE;
           connectedBlockFound=false;
           if(!tree_counter[block2].empty())
                tree_counter[block1].merge(tree_counter[block2]);
           else
                tree_counter[block1].push_back(block2);
           tree_counter[block2].push_back(0);
           tree_counter[block2].push_back(block1);
           continue;
        }
    }
    e->EdgeType=CurrentSource;
    ++_SimplifiedEdges;
  }
  free(nodeBlockA);
  free(nodeCovered);
}

void FlowGraph::replaceOffEdges()
{
  edge e;
  node startNode=nodePtrList[0];
  node n1,n2;
  int* nodeBlockA = (int*) calloc(_Nodes,sizeof(int));
  bool* nodeCovered = (bool*) calloc(_Nodes,sizeof(bool));
  queue <node> queueOfNodes;
  int counter=1;
  bool treeEdgeFound,connectedBlockFound;
  list<int> tree_counter[_Nodes];
  int block1,block2;
  forall_edges(e)
	e->edgeCovered=BLANK;
  forall_nodes(startNode)
  {
     if(nodeCovered[startNode->nodeNumber]==true) continue;
     if(startNode->edgeList.empty()) continue;
     queueOfNodes.push(startNode);
     treeEdgeFound=false;
     //This part builds tree of saturated and unsaturated edges(not forest)
     while(!queueOfNodes.empty())
     {
       n1=queueOfNodes.front();
       forall_adj_edges(e,n1)
       {
        n2= other_node(e,n1);
        if(n1==n2) continue;
        if(e->EdgeClass != OFF && e->EdgeType!=CurrentSource)
        {
          if(!nodeCovered[n2->nodeNumber])
          {
            e->edgeCovered = TREE;
            nodeCovered[n2->nodeNumber]=true;
            nodeBlockA[n2->nodeNumber] = counter;
            treeEdgeFound=true;
            queueOfNodes.push(n2);
          }
        }
       }
       queueOfNodes.pop();
     }
     if(treeEdgeFound)
     {
       nodeCovered[startNode->nodeNumber] =true;
       nodeBlockA[startNode->nodeNumber] = counter;
       ++counter;
     }
  }
  for(int i=0; i<_Edges;++i)
  {
    e=edgePtrList[i];
    if(e->EdgeClass!=OFF)
        continue;
    n1=e->Source;
    n2=e->Sink;

    if(!nodeCovered[n2->nodeNumber] &&!nodeCovered[n1->nodeNumber])
    {
        e->edgeCovered = TREE;
        nodeBlockA[n2->nodeNumber]=counter;
        nodeBlockA[n1->nodeNumber]=counter;
        nodeCovered[n1->nodeNumber] =true;
        nodeCovered[n2->nodeNumber] =true;
        ++counter;
        continue;
    }
    if(!nodeCovered[n2->nodeNumber]){
        e->edgeCovered = TREE;
        nodeBlockA[n2->nodeNumber]=nodeBlockA[n1->nodeNumber];
        nodeCovered[n2->nodeNumber] =true;
        continue;
    }
    if(!nodeCovered[n1->nodeNumber]){
        e->edgeCovered = TREE;
        nodeBlockA[n1->nodeNumber]=nodeBlockA[n2->nodeNumber];
        nodeCovered[n1->nodeNumber] =true;
        continue;
    }
    if(nodeBlockA[n2->nodeNumber]!=nodeBlockA[n1->nodeNumber])
    {
        connectedBlockFound=false;
        block1=nodeBlockA[n2->nodeNumber]<nodeBlockA[n1->nodeNumber]? nodeBlockA[n2->nodeNumber]:nodeBlockA[n1->nodeNumber];
        block2=nodeBlockA[n2->nodeNumber]<nodeBlockA[n1->nodeNumber]? nodeBlockA[n1->nodeNumber]:nodeBlockA[n2->nodeNumber];
        list<int> :: iterator p;
        while(!tree_counter[block2].empty()){
           p=tree_counter[block2].begin();
           if(*p==0)
                if(block1<*(++p))
                        block2=*p;
                else{
                        block2=block1;
                        block1=*p;
                }
           else
                break;
        }
        while(!tree_counter[block1].empty()){
           p=tree_counter[block1].begin();
           if(*p==0)
                block1=*(++p);
           else
                break;
        }
        for(p=tree_counter[block1].begin(); p!=tree_counter[block1].end(); ++p)
           if(*p==block2){
                connectedBlockFound =true;
                break;
           }
        if(block2==1 || block1==block2)
           continue;
        if(!connectedBlockFound){
           e->edgeCovered = TREE;
           connectedBlockFound=false;
           if(!tree_counter[block2].empty())
                tree_counter[block1].merge(tree_counter[block2]);
           else
                tree_counter[block1].push_back(block2);
           tree_counter[block2].push_back(0);
           tree_counter[block2].push_back(block1);
           continue;
        }

    }
    e->EdgeType=CurrentSource;
    ++_SimplifiedEdges;
  }
  free(nodeBlockA);
  free(nodeCovered);

 /* for(int i=0; i<_Edges;++i)
  {
    e=edgePtrList[i];
    cout<<e->EdgeType<<" "<<e->Source->nodeNumber+1<<" "<<e->Sink->nodeNumber+1<<" "<<e->EdgeCost<<"  "<<e->EdgeCapacity<<" "<<e->EdgeCurrent<<" "<<e->EdgeClass<<" "<<e->edgeCovered<<endl;
  }*/
}
 
void FlowGraph::buildNewProblem(char *output)
{
  edge e;
  char fileName[80];
  strcpy(fileName,output);
  strcat(fileName,".mdf");
  fstream out(fileName,ios::out);
  out<<"c Modified MinCostFlow problem"<<endl;
  out<<"p min "<<_Nodes<<" "<<_Edges<<" "<<_SimplifiedEdges<<endl;
  out<<"n "<<flowSource<<" "<<nodeCurrent<<endl;
  out<<"n "<<flowSink<<" "<<-nodeCurrent<<endl;
  for(int i=0; i<_Edges;++i)
  {
    e=edgePtrList[i];
    if(e->isEdgeType(VoltageSource))
       out<<e->EdgeType<<" "<<e->Source->nodeNumber+1<<" "<<e->Sink->nodeNumber+1<<" "<<e->EdgeCapacity<<" "<<e->EdgeCost<<" "<<endl;
    else if(e->isEdgeType(CurrentSource))
       if(e->isEdgeClass(UNSAT))
        out<<e->EdgeType<<" "<<e->Source->nodeNumber+1<<" "<<e->Sink->nodeNumber+1<<" "<<round(e->EdgeCurrent)<<" "<<e->EdgeCapacity<<" "<<e->EdgeCost<<endl;
       else if(e->isEdgeClass(OFF))
        out<<e->EdgeType<<" "<<e->Source->nodeNumber+1<<" "<<e->Sink->nodeNumber+1<<" 0 "<<e->EdgeCapacity<<" "<<e->EdgeCost<<endl;
       else
        out<<e->EdgeType<<" "<<e->Source->nodeNumber+1<<" "<<e->Sink->nodeNumber+1<<" "<<e->EdgeCapacity<<" "<<e->EdgeCapacity<<" "<<e->EdgeCost<<endl;
    else
       out<<e->EdgeType<<" "<<e->Source->nodeNumber+1<<" "<<e->Sink->nodeNumber+1<<" 0 "<<e->EdgeCapacity<<" "<<e->EdgeCost<<endl;
  }
}
  
