#ifndef _GRAPH_H
#include "Graph.h"
#endif

#ifndef _GLOBAL_H
#include "../Include/global.h"
#endif

void Graph::setGraphParameters(Int nodes, Int edges)
{
  numberOfNodes = nodes;
  numberOfEdges = edges;
}
  
void Graph::prepareAdjacencyOfCircuit(Int branchNumber, Int sourceNode,
	                                           Int sinkNode, Int branchValue)
{
  edgePtrMap[branchNumber] = new Edge; //creates an edge

  //sets the data of source Node
  if(nodePtrMap[sourceNode]==NULL)
  {
	 nodePtrMap[sourceNode]=new Node;
	 nodePtrMap[sourceNode]->setNodeNumber(sourceNode);
  }
  nodePtrMap[sourceNode]->setAdjacencyOfNode(edgePtrMap[branchNumber]);

  //sets the data of sink Node
  if(nodePtrMap[sinkNode]==NULL)
  {
	 nodePtrMap[sinkNode]=new Node;
	 nodePtrMap[sinkNode]->setNodeNumber(sinkNode);
  }
  nodePtrMap[sinkNode]->setAdjacencyOfNode(edgePtrMap[branchNumber]);

  // set the edge now
  // cout << "branchValue is " << branchValue << endl;

  edgePtrMap[branchNumber]->setEdgeData(branchNumber,branchValue);
  edgePtrMap[branchNumber]->setNodeInfo(nodePtrMap[sourceNode],nodePtrMap[sinkNode]);
  //cout << "exiting from prepareAdjacencyOfCircuit" << endl;
}

void Graph::printGraph()
{

  clock_t init=clock();

  if(nodePtrMap.size()==0)
  {
	cout<<"Graph doesn't exist.Exiting safely " << endl;
	return;
  }


  map<Int,Node*>:: iterator itr1 = nodePtrMap.begin();
  while(itr1 != nodePtrMap.end())
  {
	(*itr1).second->showNodeStatistics(0,0);
	//(*itr1).second->showNodeStatistics(0,1);
	itr1++;
  }

  clock_t final=clock();
  timeInPrintingGraph=Double((final-init)/CLOCKS_PER_SEC);
  cout << "Exiting safely" << endl;
}

void Graph::deleteGraph()
{
  clock_t init=clock();
  // Clears Node Map

  if(nodePtrMap.size()==0 && edgePtrMap.size()==0)
  {
	cout<<"Graph doesn't exist.Exiting safely " << endl;
	return;
  }

  map<Int,Node*>:: iterator itr1 = nodePtrMap.begin();
  while(itr1 != nodePtrMap.end())
  {
	(*itr1).second->removeNodeAdjacency();
	delete (*itr1).second;
	itr1++;
  }
  nodePtrMap.clear();

  // Clears Edge Map
  /* for(UnsignInt i=0;i<edgePtrMap.size();i++)
	 if(edgePtrMap[i]!=NULL)
	 delete edgePtrMap[i]; */

  map<Int,Edge*>:: iterator itr2 = edgePtrMap.begin();
  while(itr2 != edgePtrMap.end())
  {
	delete (*itr2).second;
	itr2++;
  }
  edgePtrMap.clear();

  clock_t final=clock();
  timeInDeletingGraph=Double((final-init)/CLOCKS_PER_SEC);
  cout << "Graph deleted" << endl;
}

void Graph::findNodePotentialsOfGraph(int* nodePotential)
{
//  clock_t init=clock();

  enum nodeColours {WHITE,GRAY,BLACK};
  map<Node*,Int> colour; // keep track of node colours

  map<Int,Node*>:: iterator itr0 = nodePtrMap.begin();
  while(itr0!=nodePtrMap.end())
  {
	colour[(*itr0).second]=WHITE;
	itr0++;
  }

  Node* initalNode=nodePtrMap[0];
  nodePotential[nodePtrMap[0]->getNodeNumber()]=0;
  //cout << nodePtrMap[0]->getNodeNumber() << endl;
  queue<Node*> queueOfNodes; // Keeps the Gray Nodes
  colour[initalNode]=GRAY;
  queueOfNodes.push(initalNode);
  Node *tempNode1,*tempNode2;
  
  
  while(!queueOfNodes.empty())
  {
	tempNode1=queueOfNodes.front();
	list<Edge*>:: iterator itr1 = (tempNode1->getEdgesIncidentOnNode()).begin();
	while(itr1 != (tempNode1->getEdgesIncidentOnNode()).end())
	{
	 if((*itr1)->getEdgeWeight()!=0)
	 {
	  if((tempNode2=(*itr1)->getSourceNode()) == tempNode1)
		tempNode2=(*itr1)->getSinkNode();
	  if(colour[tempNode2]==WHITE)
	  {
		//cout << (*itr1)->getEdgeWeight() << endl;
		if(tempNode2 == (*itr1)->getSourceNode())
		  nodePotential[tempNode2->getNodeNumber()]=nodePotential[tempNode1->getNodeNumber()] + (*itr1)->getEdgeWeight();
		else
		  nodePotential[tempNode2->getNodeNumber()]=nodePotential[tempNode1->getNodeNumber()] - (*itr1)->getEdgeWeight();

		colour[tempNode2]=GRAY;
		queueOfNodes.push(tempNode2);
		//printf("nodePotential[%d] %0.20lf\n",tempNode2->getNodeNumber(),nodePotential[tempNode2]);
		//printf("nodePotential %d \n",nodePotential[tempNode2->getNodeNumber()]);
	  }
	 }
	  itr1++;
	}
	queueOfNodes.pop();
	colour[tempNode1]=BLACK;
  }
  
  FILE* fpSPG = fopen("nodePotentialOfShPaGph","w+");
  fprintf(fpSPG,"%d %d\n",numberOfNodes,2*numberOfEdges);
  for(int i=0;i<int(nodePtrMap.size());i++)
	fprintf(fpSPG,"%d %d\n",i,abs(nodePotential[i]));
  
  fclose(fpSPG);
  
  // Frees memory
  colour.clear();
//  clock_t final=clock();
}
