#include<string.h>
#include<stdlib.h>
#include<stdio.h>
#include<math.h>
#include<time.h>
#include<ctype.h>
#include<iostream>

#ifndef _CONVERTER_H
#include "converter.h"
#endif

#ifndef _GRAPH_H
#include "Graph.h"
#endif

void converter::readGraph(Char* circuitName)
{

  Int treeBranch=0;
  Int branchNumber=0,sourceNode=0,sinkNode=0;
  int value=0;
  
  FILE* from;
  from=fopen(circuitName,"r+");

  fscanf(from,"%d %d %d\n",&numberOfNodes,&numberOfBranches,&treeBranch);

  for(int i=0;i<numberOfBranches;i++)
  {
	if(i+1<=treeBranch)
	{
	  fscanf(from,"%d %d %d %d\n",&branchNumber,&sourceNode,&sinkNode,&value);
	  networkGraph.prepareAdjacencyOfCircuit(branchNumber,sourceNode,sinkNode,value);
	}
	else
	{
	  fscanf(from,"%d %d %d\n",&branchNumber,&sourceNode,&sinkNode);
	  networkGraph.prepareAdjacencyOfCircuit(branchNumber,sourceNode,sinkNode,0);
	}
  }
 
  fclose(from);
  networkGraph.setGraphParameters(numberOfNodes, numberOfBranches);
  
  return;
}

void converter::convertGraphIntoElectricalCkt(char* filename,int hint)
{
  clock_t init_time = clock(); 
  readGraph(filename);
  //networkGraph.printGraph();

  int* nodePotential = new int[numberOfNodes];
  for(int i=0;i <numberOfNodes; i++)
	nodePotential[i]=0;

  networkGraph.findNodePotentialsOfGraph(nodePotential);

  //cout << "Calculated Node potentials are : " << endl;
  //for(int i=0;i <numberOfNodes; i++)
  //	cout << nodePotential[i] << endl;

  calculateCostAndGenerateElectricalCkt(filename,nodePotential,hint);
 
  delete [] nodePotential;
  
  clock_t final_time = clock(); 
  timeInGraphToCircuitConversion = double(final_time-init_time)/CLOCKS_PER_SEC;
}

void converter::calculateCostAndGenerateElectricalCkt(char* fileName,int* nodePotential, int hint)
{
  Int treeBranch=0;
  Int branchNumber=0,sourceNode=0,sinkNode=0;
  int value=0;
  
  FILE* from = fopen(fileName,"r+");
  FILE* toGTform=fopen("gtFormatShGraph","w+");
  FILE* toDIMACSform=fopen("dimacsFormatShGraph","w+");
  
  fscanf(from,"%d %d %d\n",&numberOfNodes,&numberOfBranches,&treeBranch);
  fprintf(toGTform,"%d %d\n",numberOfNodes,2*numberOfBranches);
  fprintf(toDIMACSform,"%c %s %d %d\n",'p', "SHP", numberOfNodes, 2*numberOfBranches);

  int indexGT=1;
  int coTreeCost=0,tempVal=0;
  //int negativeCostBranches=int(0.01*2*numberOfBranches);
  //cout << "negativeCostBranches " << negativeCostBranches << endl;
  
  for(int i=0;i<numberOfBranches;i++)
  {
	if(i+1<=treeBranch)
	{
	  fscanf(from,"%d %d %d %d\n",&branchNumber,&sourceNode,&sinkNode,&value);
	  
	  fprintf(toGTform,"%d %d %d %d\n",indexGT,sourceNode,sinkNode,value);
	  indexGT++;
	  
	  fprintf(toDIMACSform,"a %d %d %d\n",sourceNode+1,sinkNode+1,value);

	  tempVal = -1*(1+rand()%value);
	  fprintf(toGTform,"%d %d %d %d\n",indexGT,sinkNode,sourceNode,tempVal);
	  indexGT++;

	  fprintf(toDIMACSform,"a %d %d %d\n",sinkNode+1,sourceNode+1,tempVal);
	
	}
	else
	{
	  fscanf(from,"%d %d %d\n",&branchNumber,&sourceNode,&sinkNode);
	  
	  coTreeCost = nodePotential[sourceNode]-nodePotential[sinkNode];	
	  //cout << "coTreeCost " << coTreeCost << endl;

	  if(coTreeCost==0)
	  {
		fprintf(toGTform,"%d %d %d %d\n",indexGT,sourceNode,sinkNode,10);
		fprintf(toDIMACSform,"a %d %d %d\n",sourceNode+1,sinkNode+1,10);
	  }
	  else
	  {
		fprintf(toGTform,"%d %d %d %d\n",indexGT,sourceNode,sinkNode,coTreeCost);
		fprintf(toDIMACSform,"a %d %d %d\n",sourceNode+1,sinkNode+1,coTreeCost);
	  }

	  indexGT++;

	  if(coTreeCost < 0)
	  {
		tempVal = abs(coTreeCost)+rand()%(abs(coTreeCost));
		fprintf(toGTform,"%d %d %d %d\n",indexGT,sinkNode,sourceNode,tempVal);
		fprintf(toDIMACSform,"a %d %d %d\n",sinkNode+1,sourceNode+1,tempVal);
	  }
	  else
	  {
		if(coTreeCost==0)
		{
		  fprintf(toGTform,"%d %d %d %d\n",indexGT,sinkNode,sourceNode,10);
		  fprintf(toDIMACSform,"a %d %d %d\n",sinkNode+1,sourceNode+1,10);
		}
		else
		{
		  tempVal = -1*(1+rand()%coTreeCost);
		  fprintf(toGTform,"%d %d %d %d\n",indexGT,sinkNode,sourceNode,tempVal);
		  fprintf(toDIMACSform,"a %d %d %d\n",sinkNode+1,sourceNode+1,tempVal);
		}
	  }
	  indexGT++;
	}
  }

  fclose(toDIMACSform);
  fclose(toGTform);
  fclose(from);

  cout << "Graph in gt and DIMACS format dumped." << endl;

  FILE* to = fopen("eqCircuit","w+");
  FILE* to1 = fopen("eqCktNoI","w+");
  from = fopen("gtFormatShGraph","r+");

  fscanf(from,"%d %d\n",&numberOfNodes,&numberOfBranches);
  int index1=numberOfNodes;
  int index2=numberOfBranches+1;

  for(int i=0;i<numberOfBranches;i++)
  {
	fscanf(from,"%d %d %d %d\n",&branchNumber,&sourceNode,&sinkNode,&value);

	fprintf(to,"D%d %d %d %s\n",branchNumber,sourceNode,index1,"DMOD1");
	fprintf(to,"V%d %d %d %d\n",index2,index1,sinkNode,value);

	fprintf(to1,"D%d %d %d %s\n",branchNumber,sourceNode,index1,"DMOD1");
	fprintf(to1,"V%d %d %d %d\n",index2,index1,sinkNode,value);

	index1++;
	index2++;
  }

  fprintf(to1,".end");
  fclose(to1);
  fclose(from);

  cout << "Equivalent circuit dumped." << endl;

  int connectNode=0;
  cout << "Which node to be connect to all other node : ";
  cin >> connectNode;

  for(int i=0;i<numberOfNodes;i++)
  {
	if(i!=connectNode)
	{
	  fprintf(to,"I%d %d %d %lf\n",index2,i,connectNode,0.01);
	  //fprintf(to,"I%d %d %d %d\n",index2,i,connectNode,0.1);
	  //fprintf(to,"I%d %d %d %d\n",index2,connectNode,i,0.1);
	  //fprintf(to,"I%d %d %d %d\n",index2,connectNode,i,1);
	  index2++;
	}
  }

  fprintf(to,".end");
  fclose(to);
}

void converter::printGraphOfNetwork(Graph& g)
{
  g.printGraph();
}

void converter::printGraphOfNetwork()
{
  networkGraph.printGraph();
}

void converter::deleteGraphOfNetwork(Graph& g)
{
  g.deleteGraph();
}
