#ifndef _SPGGEN_H
#include "spgGen.h"
#endif

#ifndef _GLOBAL_H
#include "../Include/global.h"
#endif


void Circuit::setParametersOfPlanarCircuit(Double nodesDesired,Double percentageVoltageSources,Double percentageCurrentSource)
{
  if(!nodesDesired)
  {
	cout << "number of nodes cann't be zero" << endl;
	exit(3);
  }
  // Computes grids sizes for the planar circuit
  nodesInArowOfInnerGrid = Int((2+sqrt(4+(8*(nodesDesired-1))))/4);
  nodesInArowOfOuterGrid = nodesInArowOfInnerGrid+1;

  // Sets circuit parameters 
  numberOfNodes = Int(pow(nodesInArowOfInnerGrid,2) + pow(nodesInArowOfOuterGrid,2));
  numberOfBranches =  Int(4*pow(nodesInArowOfInnerGrid,2) + 2*nodesInArowOfOuterGrid * nodesInArowOfInnerGrid);
  numberOfVoltageSources = Int(Double(numberOfBranches)*percentageVoltageSources/100);
  numberOfCurrentSources = Int(Double(numberOfBranches)*percentageCurrentSource/100);
  numberOfResistances = numberOfBranches-(numberOfCurrentSources+numberOfVoltageSources);

  if(numberOfVoltageSources + numberOfCurrentSources + numberOfResistances < numberOfBranches)
	numberOfResistances += numberOfBranches-(numberOfResistances+numberOfCurrentSources+
		numberOfVoltageSources);

  //  cout << numberOfVoltageSources << " " << numberOfCurrentSources << " " << numberOfResistances << endl;

}

void Circuit::showPlanarCircuitParameters()
{
  // Verify the circuit parameters computed 

  cout << "Size of outer grid:: " << nodesInArowOfOuterGrid << " Size of inner grid:: " << nodesInArowOfInnerGrid
	<< " Total Nodes:: " << numberOfNodes << " Total Branches:: " << numberOfBranches  << endl;
}

void Circuit::generatePlanarCircuit()
{
  // seeds the random number generator
  srand((unsigned int)time(NULL)); 

  // Defines the tree for the circuit
  Int index1=0;
  Int totalNumberOfTreeBranches = numberOfNodes-1;
  Int **treeBranch=new Int*[totalNumberOfTreeBranches];
  //  cout << "toal number of tree branches is " << totalNumberOfTreeBranches << endl;
  // now fills entry 
  for(Int i=0;i<(nodesInArowOfOuterGrid*nodesInArowOfOuterGrid)-1;i += nodesInArowOfOuterGrid)
  {
	//cout << "i is " << i << endl;
	for(Int k=0;k<nodesInArowOfOuterGrid-1;k++)
	{
	  treeBranch[i+k] = new Int[2];

	  treeBranch[i+k][0]=i+k;
	  treeBranch[i+k][1]=i+k+1;
	  //	  cout << i+k << " " << treeBranch[i+k][0] << " " << treeBranch[i+k][1] << endl;
	  //	   cout << "generated " << i+k << endl;
	}
	// Sets the edge between node "i" and "i+4"
	if(index1<nodesInArowOfOuterGrid-1)
	{
	  treeBranch[i+nodesInArowOfOuterGrid-1] = new Int[2];

	  treeBranch[i+nodesInArowOfOuterGrid-1][0]=i;
	  treeBranch[i+nodesInArowOfOuterGrid-1][1]=i+nodesInArowOfOuterGrid;
	  //	  cout << i+(nodesInArowOfOuterGrid-1) << " " << treeBranch[i+nodesInArowOfOuterGrid-1][0] 
	  //		   << " " << treeBranch[i+nodesInArowOfOuterGrid-1][1] << endl;
	  // cout << "generated " << i+(nodesInArowOfOuterGrid-1) << endl;
	  index1++;
	}
  }

  // Now it advances the tree generation by attaching nodes of the inner grid to the
  // existing tree
  index1=0;
  Int index2=0;
  Int nodesCovered = nodesInArowOfOuterGrid*nodesInArowOfOuterGrid;
  //  cout << "nodesCovered is " << nodesCovered << endl;

  for(Int i=0;i<nodesInArowOfInnerGrid*nodesInArowOfInnerGrid;i++)
  {
	treeBranch[(nodesCovered-1)+i] =  new Int[2];

	treeBranch[(nodesCovered-1)+i][0]=index1;
	treeBranch[(nodesCovered-1)+i][1]=nodesCovered+i;

	//	cout << (nodesCovered-1) + i << " " << treeBranch[(nodesCovered-1)+i][0] 
	//	     << " " << treeBranch[(nodesCovered-1)+i][1] << endl;
	//  cout << "i is " << i << " and index1 is " << index1 << endl;
	index2++;  
	if(index2 == nodesInArowOfInnerGrid)
	{
	  index1 += 2;
	  index2=0;
	}
	else
	  index1++;
  }

  cout << "tree generated" << endl;
  // Does the testing here for the tree generated
  //for(Int i=0;i<totalNumberOfTreeBranches;i=i+1)
  //	cout << i  << " " << treeBranch[i][0] << " " << treeBranch[i][1] << endl;

  // Here it selects the voltage source locations in the tree and put them
  // in a file named as circuit 

  Int resistanceValue=0;
  //Int resistanceValue=1;

  ofstream to("spGraph");
  if(!to) {cerr << "Can't open file for writing!" << endl; exit(3);} 

  to << numberOfNodes << " " << numberOfBranches 
	 << " " << totalNumberOfTreeBranches << endl; 

  cout << "Total no of tree branches is "<< totalNumberOfTreeBranches << endl;

  // Prints the branch numbers chosen as voltage source.
  //for(Int k=0;k<numberOfVoltageSources;k++)
  //	cout << indexListOfVoltageSources[k] << " ";
  //cout << endl;

  index1=1; // Resets index1 to 1. index1 is used as a counter to
  // the number of devices used in the circuit. It will be 
  // used for this from now onwards.
  for(Int k=0;k<totalNumberOfTreeBranches;k++)
  {
	resistanceValue = rand()%MaxPermissibleValueOfResistance;
	if(!resistanceValue) resistanceValue=10;
	to << index1 << " " << treeBranch[k][0] << " " << treeBranch[k][1] 
	   << " " << resistanceValue << endl;
	index1++;
  }

  // here it clears the memory

  for(Int i=0;i<totalNumberOfTreeBranches;i++)
	delete[] treeBranch[i];
  delete[] treeBranch;

  // Now it completes the circuit.

  Int resistancesLeftToPutInCircuit=numberOfResistances-(totalNumberOfTreeBranches-numberOfVoltageSources);

  cout << "BranchesLeftToPutInGraph " << resistancesLeftToPutInCircuit << endl;
  Int sourceNode=0,sinkNode=0;
  index1=totalNumberOfTreeBranches+1; // Initialises index1

  // First it connects the nodes of degree One so that they form a complete circuit
  // In this sequence, it connects the nodes having degree one of the outer grid 

  for(Int i=0;i<nodesInArowOfOuterGrid-1;i++)
  {
	if(!(rand()%2))
	{
	  sourceNode=(i+1)*nodesInArowOfOuterGrid-1;
	  sinkNode=(i+2)*nodesInArowOfOuterGrid-1;
	}
	else
	{
	  sourceNode=(i+2)*nodesInArowOfOuterGrid-1;
	  sinkNode=(i+1)*nodesInArowOfOuterGrid-1;
	}

	resistanceValue = rand()%MaxPermissibleValueOfResistance;
	if(!resistanceValue) resistanceValue=10;
	//cout << " index1 is "<< index1 << endl; 
	to << index1 << " " << sourceNode << " " << sinkNode << endl;

	index1++;
	resistancesLeftToPutInCircuit--;
  }
  // Here it connects nodes of the inner grid to the nodes of the 
  // outer grid which form windows in which nodes of inner grid 
  // are placed.
  index2=1; // it is used as an index to the node where 
  // the nodes of the inner grid will be attached.
  Int index3=0; // used as a formfeed
  for(Int i=0;i<nodesInArowOfInnerGrid*nodesInArowOfInnerGrid;i++)
  {
	if(!(rand()%2))
	{
	  sourceNode=nodesInArowOfOuterGrid*nodesInArowOfOuterGrid+i;
	  sinkNode=index2;
	}
	else
	{
	  sourceNode=index2;
	  sinkNode=nodesInArowOfOuterGrid*nodesInArowOfOuterGrid+i;
	}

	resistanceValue = rand()%MaxPermissibleValueOfResistance;
	if(!resistanceValue) resistanceValue=10;
	to <<  index1 << " " << sourceNode << " " << sinkNode << endl;

	index1++;
	index3++;
	resistancesLeftToPutInCircuit--;
	if(index3==nodesInArowOfOuterGrid-1)
	{
	  index3=0;
	  index2+=2;
	}
	else
	  index2++;
  }
  cout << "Branches left to put in Graph is now " << resistancesLeftToPutInCircuit << endl; 
  // Here it addes resistances in the columns to the circuit again. If 
  // resistances end up in between then it pads the two nodes with the
  // current sources. Since the circuit is already connected by this time, 
  // addition of current sources wont make any harm

  Int currentSourceValue=0;
  for(Int i=0;i<nodesInArowOfOuterGrid-2;i++)
  {
	for(Int j=0;j<nodesInArowOfOuterGrid-1;j++)
	{
	  if(resistancesLeftToPutInCircuit!=0)
	  {
		if(!(rand()%2))
		{
		  sourceNode=(nodesInArowOfOuterGrid*j)+i+1;
		  sinkNode= (nodesInArowOfOuterGrid*(j+1))+i+1;
		}
		else
		{
		  sourceNode=(nodesInArowOfOuterGrid*(j+1))+i+1;
		  sinkNode=(nodesInArowOfOuterGrid*j)+i+1;
		}
		resistanceValue = rand()%MaxPermissibleValueOfResistance;
		if(!resistanceValue) resistanceValue=10;
		to << index1 << " " << sourceNode << " " << sinkNode << endl;
	  
		index1++;
		resistancesLeftToPutInCircuit--;
	  }
	  else if (numberOfCurrentSources!=0 && resistancesLeftToPutInCircuit==0)
	  {
		if(!(rand()%2))
		{
		  sourceNode=(nodesInArowOfOuterGrid*j)+i+1;
		  sinkNode= (nodesInArowOfOuterGrid*(j+1))+i+1;
		}
		else
		{
		  sourceNode=(nodesInArowOfOuterGrid*(j+1))+i+1;
		  sinkNode=(nodesInArowOfOuterGrid*j)+i+1;
		}
		currentSourceValue = rand()%MaxPermissibleValueOfCurrentSource;
		if(!currentSourceValue) currentSourceValue=10;
		to << index1 << " " << sourceNode << " " << sinkNode << endl;
		index1++;
	  }
	}
  }
 
  // Now it connects all the inner grid nodes to the nodes of the windows 
  // in which they sit 

  index2=0; index3=0;
  for(Int i=0;i<2;i++) // "i<2" becoz we have to put two branches per window only 
  {
	for(Int j=0;j<nodesInArowOfInnerGrid*nodesInArowOfInnerGrid;j++)
	{
	  if(resistancesLeftToPutInCircuit!=0)
	  {
		if(!(rand()%2))
		{
		  sourceNode=nodesInArowOfOuterGrid+i+index2;
		  sinkNode= (nodesInArowOfOuterGrid*nodesInArowOfOuterGrid)+j;
		}
		else
		{
		  sourceNode=(nodesInArowOfOuterGrid*nodesInArowOfOuterGrid)+j;
		  sinkNode=nodesInArowOfOuterGrid+i+index2;
		}
		resistanceValue = rand()%MaxPermissibleValueOfResistance;
		if(!resistanceValue) resistanceValue=10;
		to << index1 << " " << sourceNode << " " << sinkNode << endl;
	  
		index1++;
		index3++;
		resistancesLeftToPutInCircuit--;
		if(index3==nodesInArowOfOuterGrid-1)
		{
		  index2 +=2;
		  index3=0;
		}
		else
		  index2++;
	  }
	  else if (numberOfCurrentSources!=0 && resistancesLeftToPutInCircuit==0)
	  {
		if(!(rand()%2))
		{
		  sourceNode=nodesInArowOfOuterGrid+i+index2;
		  sinkNode= (nodesInArowOfOuterGrid*nodesInArowOfOuterGrid)+j;
		}
		else
		{
		  sourceNode=(nodesInArowOfOuterGrid*nodesInArowOfOuterGrid)+j;
		  sinkNode=nodesInArowOfOuterGrid+i+index2;
		}
		currentSourceValue = rand()%MaxPermissibleValueOfCurrentSource;
		if(!currentSourceValue) currentSourceValue=10;
		to << index1 << " " << sourceNode << " " << sinkNode << endl;
		
		index1++;
		index3++;
		if(index3==nodesInArowOfOuterGrid-1)
		{
		  index2 +=2;
		  index3=0;
		}
		else
		  index2++;
	  }
	}
	index2=0;
  }
  
  // gives some finishing touch
 // to << ".tran 1ms 10ms"<< endl;
  to << ".end" << endl;
  to.close();
  cout << "Random planar graph for experimentation of all pair shortest path generated."<< endl;
}
