// GT :: Implemented gt Gaurav Trivedi
#include<GTL/graph.h>
#include<GTL/debug.h>
#include<GTL/planarity.h>
#include<GTL/bellman_ford.h>
#include<vector>
#include<iostream>
#include<fstream>
#include<stdio.h>
#include<stdlib.h>
#include<time.h>

int main(int argc,char** argv)
{
  if(argc!=3)
  {
  	cout << "Insufficient args!" << endl;
	cout << "<executable> <graph-file-name> <node>" << endl;
	exit(1);
  }

  int totalNodes=0,totalEdges=0;
  int bNumber=0,sourceNode=0,sinkNode=0,bValue=0;
  int* costArray=NULL;
    
  FILE* from;
  from = fopen(argv[1],"r+");

  fscanf(from,"%d %d\n",&totalNodes,&totalEdges);
  cout << "nodes : " << totalNodes << " edges : " << totalEdges << endl; 

  if(atoi(argv[2])>=totalNodes)
  {
	cout << "Node " << atoi(argv[2]) 
	  << " asked is not in scope. Max Node number is " 
	  << totalNodes-1 << endl;
	exit(2);
  }

  costArray = new int[totalEdges];

  graph G;
  G.make_directed();
  
  node* nodePtrArray = new node[totalNodes];
  edge* edgePtrArray = new edge[totalEdges];
  
  for(int i=0;i<totalNodes;i++)
  	nodePtrArray[i] = G.new_node();

  for(int i=0;i<totalEdges;i++)
  {
	fscanf(from,"%d %d %d %d\n",&bNumber,&sourceNode,&sinkNode,&bValue);
	edgePtrArray[i] = G.new_edge(nodePtrArray[sourceNode],nodePtrArray[sinkNode]);
	costArray[i]=bValue;
  }

  fclose(from);

  edge_map<double> w(G);

  for(int i=0; i<totalEdges;i++)
  	w[edgePtrArray[i]] = double(costArray[i]);

  node_map<int> result(G);
  node_map<node> preds(G);

  bellman_ford B;
  B.store_preds(true);
  B.source(nodePtrArray[atoi(argv[2])]);
  B.weights(w);
  
  clock_t start=clock();

  if (B.check(G) == algorithm::GTL_OK) 
  {
	cout << "check OK" << endl; 
  } 
  else 
  {
	cout << "check FAILED" << endl;
	exit(1);
  }

  if (B.run(G) == algorithm::GTL_OK) 
  {
	cout << "run OK" << endl; 
  } 
  else 
  {
	cout << "run FAILED" << endl;
	exit(2);
  }
   
  if (B.negative_cycle()) 
  {
	cout << "Negative Cycle Found" << endl;
	exit(3);
  }
  else 
  {
	cout << "No Negative Cycle" << endl;	
  }

  clock_t stop=clock();

  cout << "Time in finding shortest path " << double(stop-start)/CLOCKS_PER_SEC << endl;

  graph::node_iterator it, end;
  
  FILE* to = fopen("gtlShPathCalcResults","w");
  int index=0;
  
  //fprintf(to,"Shortest Path Calculating from %d\n",atoi(argv[2]));
  fprintf(to,"%d %d %d\n",totalNodes,totalEdges,atoi(argv[2]));
  
  for (it = G.nodes_begin(), end = G.nodes_end(); it != end; ++it)
  {
	result[*it] = int(B.distance(*it));
	//cout << result[*it] << endl;
	if(index != atoi(argv[2]))
	  fprintf(to,"%d %d\n",index,result[*it]);
	index++;
  }
  fclose(to);

  for (it = G.nodes_begin(), end = G.nodes_end(); it != end; ++it)
	preds[*it] = B.predecessor_node(*it);

  cout << "Success:: Shortest Path Distances calculated." << endl;
  G.clear();
}
